#pragma once
#ifndef CONCURRENT_QUEUE_H
#define CONCURRENT_QUEUE_H

#include <condition_variable>
#include <mutex>
#include <queue>
#include <thread>

template <typename T>
class ConcurrentQueue
{
public:
    T Pop()
    {
        std::unique_lock<std::mutex> lock(mutex_);
        while (queue_.empty())
        {
            cond_.wait(lock);
        }
        auto item = queue_.front();
        queue_.pop();
        return item;
    }

    // provided for strong exception safety
    void Pop(T & item)
    {
        std::unique_lock<std::mutex> lock(mutex_);
        while (queue_.empty())
        {
            cond_.wait(lock);
        }
        item = queue_.front();
        queue_.pop();
    }

    // if the queue is empty, just return instead of waiting
    void PopIfNonEmpty(T & item)
    {
        std::unique_lock<std::mutex> lock(mutex_);
        if (!queue_.empty())
        {
            item = queue_.front();
            queue_.pop();
        }
    }

    void Push(const T & item)
    {
        std::unique_lock<std::mutex> lock(mutex_);
        queue_.push(item);
        lock.unlock();
        cond_.notify_one();
    }

    void Push(T && item)
    {
        std::unique_lock<std::mutex> lock(mutex_);
        queue_.push(std::move(item));
        lock.unlock();
        cond_.notify_one();
    }

    // this is NOT thread-safe, but is useful if the queue is initialized in the beginning and
    // no threads can push to the queue
    bool IsEmpty() const
    {
        return queue_.empty();
    }

    std::size_t Size() const
    {
        return queue_.size();
    }

private:
    std::queue<T> queue_;
    std::mutex mutex_;
    std::condition_variable cond_;
};

#endif