#include <PhotonMapping.h>
#include <Sampling.h>
#include <RayTracer.h>
#include <chrono>
#include <random>
#include <fstream>
#include <Serialization.h>
#include <lights.h>

extern LightList lights;
extern Node rootNode;
extern int photonReductionRatio;
extern float photonMapFilterRadius;

struct ExternalPhoton
{
    ExternalPhoton() { }
    ExternalPhoton(const Photon & photon)
    {
        pos[0] = photon.pos.x;
        pos[1] = photon.pos.y;
        pos[2] = photon.pos.z;
        dir[0] = photon.dir.x;
        dir[1] = photon.dir.y;
        dir[2] = photon.dir.z;
        power[0] = photon.col.r;
        power[1] = photon.col.g;
        power[2] = photon.col.b;
    }
    Photon ToPhoton() const
    {
        Photon p;
        p.pos.x = pos[0];
        p.pos.y = pos[1];
        p.pos.z = pos[2];
        p.dir.x = dir[0];
        p.dir.y = dir[1];
        p.dir.z = dir[2];
        p.col.r = power[0];
        p.col.g = power[1];
        p.col.b = power[2];
        return p;
    }
    float pos[3]; // photon position
    float dir[3]; // incoming direction
    float power[3]; // photon power
};

void WriteExternalPhoton(std::ofstream & os, const ExternalPhoton & photon)
{
    WritePrimitive(os, photon.pos);
    WritePrimitive(os, photon.dir);
    WritePrimitive(os, photon.power);
}

void ReadExternalPhoton(std::ifstream & os, const ExternalPhoton & photon)
{
    ReadPrimitive(os, photon.pos);
    ReadPrimitive(os, photon.dir);
    ReadPrimitive(os, photon.power);
}

void PhotonMapping::InitPhotonMaps()
{
    // compute the volume of the scene
    float vol = rootNode.GetChildBoundBox().Volume(); // WARNING: this is ok... for now
    radiusHint_ = pow(8.0f * vol / numPhotonsEmitted_, 1.0f / 3.0f);
    float reductionRadiusHint = pow(8.0f * vol / (numPhotonsEmitted_ / photonReductionRatio), 1.0f / 3.0f);
    globalMap_.reset(new PhotonMap(radiusHint_));
    causticMap_.reset(new PhotonMap(radiusHint_));
    //filteredMap_.reset(new PhotonMap(reductionRadiusHint));
    std::cout << "radius hint = " << radiusHint_ << std::endl;
}
#define COMPARE(a,b) (_stricmp(a,b)==0)
void PhotonMapping::GeneratePhotonMap()
{
    AccumulateLightPower();
    InitPhotonMaps();
    std::cout << "light cache size = " << lightCache_.size() << std::endl;
    for (unsigned int i = 0u; i < numPhotonsEmitted_; ++i)
    {
        if (i % 100000 == 0) std::cout << i << "p ";
        // select a light source
        const auto light = SelectLight();
        // shoot a photon in a random direction in the sphere
        Ray photonRay;
        //if (COMPARE(light->GetName(), "sunLight"))
        //{
        //    // select a random point from light cache
        //    int j = RandInt() % lightCache_.size();
        //    photonRay = static_cast<const DirectAreaLight *>(light)->RandomPhoton(lightCache_[j]);
        //}
        //else
        {
            photonRay = light->RandomPhoton();
        }
        Color li = light->GetPhotonIntensity();
        Color photonColor = (accumLightPower_.back() * li) / (numPhotonsEmitted_ * li.Gray());
        // trace this photon
        TracePhoton(photonRay, photonColor);
    }
}

void PhotonMapping::AccumulateLightPower()
{
    accumLightPower_.resize(lights.size());
    Color accum = Color().Black();
    for (std::size_t i = 0; i < lights.size(); ++i)
    {
        accum += lights[i]->GetPhotonIntensity();
        accumLightPower_[i] = accum.Gray();
    }
}

const Light * PhotonMapping::SelectLight() const
{
    // generate a random number between 0 and total intensity
    auto randVal = RandFloat() * accumLightPower_.back();

    // find the first light source whose value is at least this random value
    for (std::size_t i = 0; i < accumLightPower_.size(); ++i)
    {
        if (accumLightPower_[i] >= randVal)
        {
            return lights[i];
        }
    }
    return 0;
}

void PhotonMapping::TracePhoton(Ray & photonRay, Color & photonColor)
{
    unsigned int diffuseBounceCount = 0;
    unsigned int causticBounceCount = 0;
    for (unsigned int i = 0; i < numPhotonBounces; ++i)
    {
        HitInfo photonHitInfo;
        // trace photon
        bool photonHit = TraceRay(photonRay, &rootNode, photonHitInfo, HIT_FRONT_AND_BACK);

        // if we hit something
        if (photonHit)
        {
            // get the hit material
            const Node * hitNode = photonHitInfo.node;
            const Material * mat = hitNode->GetMaterial();

            // store the photon if necessary
            if (mat->IsPhotonSurface() && photonHitInfo.front) // only store on diffuse surface and on the front face
            {// store the photon
                if (diffuseBounceCount == 0 && causticBounceCount > 0)
                {
                    causticMap_->insert(Photon(photonRay, photonColor, photonHitInfo));
                }
                else if ((i > 0) || (i == 0 && storeFirstHit))
                {
                    globalMap_->insert(Photon(photonRay, photonColor, photonHitInfo));
                }
            }

            // bounce the photon
            auto transportType = mat->RandomPhotonBounce(photonRay, photonColor, photonHitInfo);
            if (transportType == Material::ABSORB)
            {
                break;
            }
            else if (transportType == Material::DIFFUSE)
            {
                ++diffuseBounceCount;
            }
            else
            {
                ++causticBounceCount;
            }
        }
        else // hit nothing
        {
            break;
        }
    }
}

inline G3D::Point3 cy2G3D(const Point3 & p)
{
    G3D::Point3 out;
    out.x = p.x;
    out.y = p.y;
    out.z = p.z;
    return out;
}

void PhotonMapping::FilterPhotonMap(const std::unique_ptr<PhotonMap> & inputMap, std::unique_ptr<PhotonMap> & outputMap, float size) const
{
    G3D::Array<Photon> neighbors;
    float denom = 1 / (PI * Square(photonMapFilterRadius));
    for (auto iter = inputMap->begin(); iter != inputMap->end(); ++iter)
    {
        int i = RandInt();
        if (i % photonReductionRatio != 0) { continue; } // store one-fourth the number of original photons
        neighbors.fastClear();
        G3D::Sphere sphere(cy2G3D(iter->pos), size);
        inputMap->getIntersectingMembers(sphere, neighbors);
        if (neighbors.size() == 0) { continue; }
        // average the position
        // average the direction
        // average the colors
        Point3 avgPos(0, 0, 0);
        Point3 avgDir(0, 0, 0);
        Color avgCol(0, 0, 0);
        for (auto & photon : neighbors)
        {
            avgPos += photon.pos;
            avgDir += photon.dir;
            avgCol += photon.col;
        }
        Photon avgPhoton;
        avgPhoton.pos = avgPos * denom;
        avgPhoton.dir = (avgDir  * denom).GetNormalized();
        avgPhoton.col = avgCol * denom;
        outputMap->insert(avgPhoton);
    }
}

void PhotonMapping::SavePhotonMap(const std::unique_ptr<PhotonMap> & map, const std::string & name) const
{
    std::ofstream fs(name, std::ofstream::binary);
    int size = map->size();
    WritePrimitive<float>(fs, radiusHint_);
    WritePrimitive<int>(fs, size);
    for (auto item = map->begin(); item != map->end(); ++item)
    {
        WriteExternalPhoton(fs, ExternalPhoton(*item));
    }
    fs.close();
}

void PhotonMapping::LoadPhotonMap(std::unique_ptr<PhotonMap> & map, const std::string & name) const
{
    std::ifstream fs(name, std::ifstream::binary);
    int size = 0;
    ReadPrimitive(fs, radiusHint_);
    ReadPrimitive(fs, size);
    map.reset(new PhotonMap(radiusHint_));
    for (auto i = 0; i < size; ++i)
    {
        ExternalPhoton extPhoton;
        ReadExternalPhoton(fs, extPhoton);
        map->insert(extPhoton.ToPhoton());
    }
    fs.close();
}

void PhotonMapping::LoadLightCache()
{
    std::ifstream ifs("light_cache.txt");
    while (ifs)
    {
        Point3 p;
        ifs >> p.x >> p.y >> p.z;
        lightCache_.push_back(p);
    }
    ifs.close();
}