#pragma warning (push, 1)
#include <materials.h>
#include <lights.h>
#pragma warning (pop)
#include <RayTracer.h>
#include <common.h>
#include <PhotonMapping.h>

extern TexturedColor background;
extern ImagePlane imagePlane;
extern Node rootNode;
extern LightList lights;
extern int bounceCountMax_;
extern PhotonMapping pm;

// trace indirect illumination
Color RayTracer::DoRayTrace2nd(const Camera & cam, Pixel & pixel, SubPixel & subpix)
{
    // prepare the ray
    Ray rayWorld = GetCameraRay(cam, cam.imgWidth, pixel.x, pixel.y, subpix.rx, subpix.ry);
    rayWorld.dir.Normalize();

    HitInfo hitInfo;
    if (TraceRay(rayWorld, &rootNode, hitInfo, HIT_FRONT))
    {
        return hitInfo.node->GetMaterial()->ShadePhoton(rayWorld, hitInfo, lights, pm.globalMap_.get(), pm.causticMap_.get(), 0);
    }
    else // hit nothing (already handled in the first pass)
    {
        return Color().Black();
    }
    return Color().Black();
}

void RayTracer::DoRayTrace(const Camera & cam, Pixel & pixel, SubPixel & subpix)
{
    // prepare the ray
    Ray rayWorld = GetCameraRay(cam, cam.imgWidth, pixel.x, pixel.y, subpix.rx, subpix.ry);
    rayWorld.dir.Normalize();

    HitInfo hitInfo;
    if (TraceRay(rayWorld, &rootNode, hitInfo, HIT_FRONT))
    {
        subpix.color = hitInfo.node->GetMaterial()->ShadePhotonDirect(rayWorld, hitInfo, lights, pm.globalMap_.get(), pm.causticMap_.get());
    }
    else // the background
    {
        cyPoint3f uvw(float(pixel.x + subpix.rx) / cam.imgWidth, float(pixel.y + subpix.ry) / cam.imgHeight, 0.0f);
        subpix.color = background.Sample(uvw);
    }
}

Ray RayTracer::GetCameraRay(const Camera & cam, int imgWidth, int x, int y, float rx, float ry)
{
    Ray ray;
    ray.p = cam.pos;
    float pixelSize = imagePlane.width / imgWidth;
    Point3 q = imagePlane.topLeft + imagePlane.right * (x + rx) * pixelSize + imagePlane.down * (y + ry) * pixelSize;
    ray.dir = q - ray.p;
    return ray;
}

bool TraceRay(const Ray & ray, const Node * node, HitInfo & hitInfo, int hitSide)
{
    // transform the ray into the local coordinate frame of the node
    Ray rayLocal = node->ToNodeCoords(ray);

    // intersect the ray with the object
    bool hit = false;
    const Object * obj = node->GetObject();

    // check with the bounding box
    if (obj != nullptr && obj->GetBoundBox().IntersectRay(rayLocal, hitInfo.z))
    {
        if (obj->IntersectRay(rayLocal, hitInfo, hitSide))
        {
            hit = true;
            hitInfo.node = node;
        }
    }

    // intersect the ray with the children
    if (node->GetNumChild() > 0 && node->GetChildBoundBox().IntersectRay(rayLocal, hitInfo.z))
    {
        for (int i = 0; i < node->GetNumChild(); ++i)
        {
            const Node * childNode = node->GetChild(i);
            if (TraceRay(rayLocal, childNode, hitInfo, hitSide))
            {
                hit = true;
            }
        }
    }

    if (hit)
    {
        // transform the normal and the hit position to the outer space
        node->FromNodeCoords(hitInfo);
    }

    return hit;
}