#pragma once
#ifndef SERIALIZATION_H_
#define SERIALIZATION_H_

#include <fstream>
#include <map>
#include <vector>

template<typename T>
inline void WritePrimitive(std::ofstream & ofs, const T & t)
{
    ofs.write((const char *)&t, sizeof(t));
}

template<typename T>
inline void ReadPrimitive(std::ifstream & ifs, T & t)
{
    ifs.read((char *)&t, sizeof(t));
}

template <typename T>
inline void WritePrimitiveContainer(std::ofstream & ofs, const T & container)
{
    WritePrimitive(ofs, container.size());
    for (auto & item : container)
    {
        WritePrimitive(ofs, item);
    }
}

template <typename T>
inline void ReadPrimitiveVector(std::ifstream & ifs, std::vector<T> & container)
{
    std::size_t size;
    ReadPrimitive(ifs, size);
    container.resize(size);
    ifs.read((char *)&container[0], size * sizeof(T));
}

template <typename T>
inline void WriteClassContainer(std::ofstream & ofs, const T & container)
{
    WritePrimitive(ofs, container.size());
    for (auto & item : container)
    {
        item.Write(ofs);
    }
}

template <typename T>
inline void ReadClassVector(std::ifstream & ifs, std::vector<T> & container)
{
    std::size_t size;
    ReadPrimitive(ifs, size);
    container.resize(size);
    for (auto & item : container)
    {
        item.Read(ifs);
    }
}

template <typename T1, typename T2>
inline void WriteMapPrimitive(std::ofstream & ofs, const std::map<T1, T2> & container)
{
    WritePrimitive(ofs, container.size());
    for (auto & item : container)
    {
        WritePrimitive(ofs, item.first);
        WritePrimitive(ofs, item.second);
    }
}

template <typename T1, typename T2>
inline void ReadMapPrimitive(std::ifstream & ifs, std::map<T1, T2> & container)
{
    std::size_t size;
    ReadPrimitive(ifs, size);
    for (std::size_t i = 0; i < size; ++i)
    {
        T1 first;
        T2 second;
        ReadPrimitive(ifs, first);
        ReadPrimitive(ifs, second);
        container[first] = second;
    }
}

template <>
inline void WritePrimitive(std::ofstream & ofs, const std::string & s)
{
    std::size_t l = s.size();
    WritePrimitive(ofs, l);
    ofs.write(s.c_str(), l);
}

template <>
inline void ReadPrimitive(std::ifstream & ifs, std::string & s)
{
    s.clear();
    std::size_t l = 0;
    ReadPrimitive(ifs, l);
    if (l == 0) return;
    std::vector<char> temp(l);
    ifs.read(&temp[0], l);
    s = std::string(temp.begin(), temp.end());
}

template <typename T>
inline void WritePrimitive(std::ofstream & ofs, const std::pair<T, T> & item)
{
    WritePrimitive(ofs, item.first);
    WritePrimitive(ofs, item.second);
}

#endif