#pragma warning(push, 1)
#include <scene.h>
#include <lights.h>
#include <materials.h>
#include <objects.h>
#include <trace.h>
#include <xmlload.h>
#include <texture.h>
#pragma warning(pop)
#include <cassert>
#include <chrono>
#include <iostream>
#include <ppl.h>
#include <cmath>
#include <sstream>
#include <vector>
#include <algorithm>
#include <ConcurrentQueue.h>
#include <Pixel.h>
#include <ppl.h>
#include <Filtering.h>
#include <AdaptiveSubdivisionTracer.h>
#include <AdaptiveSamplingTracer.h>
#include <common.h>
#include <G3D/Vector3.h>
#include <PhotonMapping.h>
#include <string>

Node rootNode;
Camera camera;
RenderImage renderImage;
MaterialList materials;
LightList lights;
Sphere theSphere;
Plane thePlane;
ObjFileList objList;

TexturedColor background;
TexturedColor environment;
TextureList textureList;

unsigned int numSamplesMin = 32;
unsigned int numSamplesMax = 32;
unsigned int numMonteCarloSamples = 256; // used during final gathering
float radianceEstimationRadius = 1.0f; // used by photon mapping
float photonMapFilterRadius = 2.0f;
PhotonMapping pm;
int photonReductionRatio = 4;

ImagePlane imagePlane;
int bounceCountMax_ = 1;
std::vector<Pixel> pixels;
ConcurrentQueue<PixelBlock> taskQueue_;

void GetImagePlane()
{
    imagePlane.height = 2.0f * camera.focaldist * tan(DegreeToRadian(camera.fov) / 2);
    imagePlane.width = camera.imgWidth * (imagePlane.height / camera.imgHeight);
    imagePlane.right = camera.dir.Cross(camera.up).GetNormalized();
    imagePlane.down = camera.dir.Cross(imagePlane.right).GetNormalized();
    Point3 center = camera.pos + camera.focaldist * camera.dir.GetNormalized();
    imagePlane.topLeft = center - imagePlane.right * imagePlane.width / 2.0 - imagePlane.down * imagePlane.height / 2.0;
}

inline float Clamp(float v)
{
    return v < 0 ? 0 : v > 255 ? 255 : v;
}

void CopyImage(const std::vector<Pixel> & image, Color24 * renderImg)
{
    const float gamma = 1.0f / 2.2f;
    for (int y = 0; y < camera.imgHeight; ++y)
    {
        for (int x = 0; x < camera.imgWidth; ++x)
        {
            int pixelId = y * camera.imgWidth + x;
            //renderImg[pixelId] = image[pixelId].color;
            renderImg[pixelId].r = (unsigned char)Clamp(255.0f * pow(image[pixelId].color.r, gamma));
            renderImg[pixelId].g = (unsigned char)Clamp(255.0f * pow(image[pixelId].color.g, gamma));
            renderImg[pixelId].b = (unsigned char)Clamp(255.0f * pow(image[pixelId].color.b, gamma));
        }
    }
}

int main(int argc, char ** argv)
{
    using namespace std;
    using namespace std::chrono;

    typedef time_point<high_resolution_clock> TimePoint;
    TimePoint start = high_resolution_clock::now();

    // Load scene
    if (argc > 2)
    {
        LoadScene(argv[1]);
        bounceCountMax_ = lexical_cast<int>(argv[2]);
    }
    else
    {
        cout << "Please provide the scene file and a bounce count" << endl;
        return 0;
    }
    assert(renderImage.GetWidth() > 0);

    //// normalize all materials
    //for (auto & mat : materials)
    //{
    //    mat->Normalize();
    //}

    if (argc > 3 && std::string(argv[3]) == "build")
    {
        //pm.LoadLightCache();
        pm.GeneratePhotonMap();
        pm.FilterPhotonMap(pm.globalMap_, pm.filteredMap_, photonMapFilterRadius);
        pm.SavePhotonMap(pm.globalMap_, "global_photon_map");
        pm.SavePhotonMap(pm.causticMap_, "caustic_photon_map");
        //pm.SavePhotonMap(pm.filteredMap_, "filtered_photon_map");
        return 0;
    }
    else if (argc > 3 && std::string(argv[3]) == "filter")
    {
        std::ifstream ifs("result.dat", std::ofstream::binary);
        ReadPrimitive(ifs, camera.imgWidth);
        ReadPrimitive(ifs, camera.imgHeight);
        std::size_t size;
        ReadPrimitive(ifs, size);
        pixels.resize(size);
        for (auto & pixel : pixels)
        {
            pixel.ReadFromStream(ifs);
        }
        ifs.close();
        FilterImage(camera.imgWidth, camera.imgHeight, pixels);
        renderImage.Init(camera.imgWidth, camera.imgHeight);
        CopyImage(pixels, renderImage.GetPixels());
        renderImage.SaveImage((string(argv[1]) + ".ppm").c_str());
        return 0;
    }
    else
    {
        pm.LoadPhotonMap(pm.globalMap_, "global_photon_map");
        //pm.LoadPhotonMap(pm.globalMap_, "filtered_photon_map");
        pm.LoadPhotonMap(pm.causticMap_, "caustic_photon_map");
    }

    GetImagePlane();
    renderImage.Init(camera.imgWidth, camera.imgHeight);
    Color24 * img = renderImage.GetPixels();
    unsigned int * sampleCountImg = renderImage.GetSampleCount();

    unsigned int numCores = thread::hardware_concurrency();
    cout << "Number of cores = " << numCores << endl;

    // initialize the pixel array
    pixels.resize(camera.imgWidth * camera.imgHeight);
    for (int y = 0; y < camera.imgHeight; ++y)
    {
        for (int x = 0; x < camera.imgWidth; ++x)
        {
            int pixID = y * camera.imgWidth + x;
            pixels[pixID].x = x;
            pixels[pixID].y = y;
        }
    }

    // spawn the threads
    vector<thread> threads(numCores);
    AdaptiveSamplingTracer tracer(camera, img, sampleCountImg);
    tracer.BuildQueue();
    for (auto & t : threads)
    {
        t = thread(tracer);
    }

    // wait
    for (auto & t : threads)
    {
        t.join();
    }

    // filter the image
    std::ofstream ofs("result.dat", std::ofstream::binary);
    WritePrimitive(ofs, camera.imgWidth);
    WritePrimitive(ofs, camera.imgHeight);
    WritePrimitive(ofs, pixels.size());
    for (auto & pix : pixels)
    {
        pix.WriteToStream(ofs);
    }
    ofs.close();
    FilterImage(camera.imgWidth, camera.imgHeight, pixels);
    CopyImage(pixels, img);

    TimePoint end = high_resolution_clock::now();
    duration<double> elapsed = end - start;
    std::cout << "Elapsed time = " << elapsed.count() << " s" << std::endl;

    //renderImage.ComputeSampleCountImage();
    renderImage.SaveImage((string(argv[1]) + ".ppm").c_str());
    //renderImage.SaveSampleCountImage((string(argv[1]) + "_sample_count.ppm").c_str());
}