#pragma warning(push, 1)
#include <materials.h>
#pragma warning(pop)
#include <trace.h>
#include <algorithm>
#include <cassert>
#include <iostream>
#include <common.h>
#include <Sampling.h>
#include <PhotonMapping.h>

extern Node rootNode;
extern TexturedColor environment;
extern int bounceCountMax_;
extern unsigned int numMonteCarloSamples;
extern float radianceEstimationRadius;

inline void ReflectRayAboutPlane(Ray & ray, const Point3 & n)
{
    ray.dir = ray.p + ray.dir + n * (2 * -ray.dir.Dot(n)) - ray.p;
}

// incoming ray is from outside to surface, and should be normalized
inline Ray ComputeRefractionRay(const Point3 & incomingDir, const Point3 & p, const Point3 & N, float n)
{
    Ray refractRay;
    refractRay.p = p;
    float dn = N.Dot(incomingDir);
    float dpn = -sqrt(std::max(1 - n * n * (1 - dn * dn), 0.0f));
    refractRay.dir = (n * incomingDir - (n * dn - dpn) * N).GetNormalized();
    assert(refractRay.dir.Dot(N) <= 0);
    return refractRay;
}

void MtlBlinn::Normalize()
{
    diffuseNorm = 1 / PI;
    float n = glossiness;
    reflectionNorm = (n + 2) * (n + 4) / (8 * PI * (n + pow(2.0f, -n / 2)));
    refractionNorm = reflectionNorm;
}

inline G3D::Point3 cy2G3D(const Point3 & p)
{
    G3D::Point3 out;
    out.x = p.x;
    out.y = p.y;
    out.z = p.z;
    return out;
}

Color MtlBlinn::ShadePhotonDirect(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap) const
{
    Color color = Color().Black();
    Color direct = DirectLighting(ray, hitInfo, lights);
    if (direct == direct)
    {
        color += direct;
    }
    //if (IsPhotonSurface()) // diffuse surface, spawn mc rays
    //{
    //    /*auto caustic = GatherFromPhotonMap(ray, hitInfo, globalMap);
    //    if (caustic == caustic) { color += caustic; }*/
    //    G3D::Sphere querySphere(cy2G3D(hitInfo.p), radianceEstimationRadius);
    //    G3D::Array<Photon> neighbors;
    //    globalMap->getIntersectingMembers(querySphere, neighbors);
    //    color = DirectLighting(ray, hitInfo, lights) * neighbors.size() / (PI * radianceEstimationRadius * radianceEstimationRadius);
    //}
    return color;
}

// IMPORTANT: assumptions:
// a surface with reflectionGlossiness > 0 or refractionGlossiness > 0 will have a diffuse > 0
Color MtlBlinn::ShadePhoton(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, int bounceCount) const
{
    Color color = Color().Black();
    if (bounceCount > bounceCountMax_) { return color; }

    unsigned int diffuseRaysCount = 0;
    unsigned int reflectRaysCount = 0;
    unsigned int refractRaysCount = 0;

    //color += DirectLighting(ray, hitInfo, lights); // direct lighting

    if (IsPhotonSurface()) // diffuse surface, spawn mc rays
    {
        auto caustic = GatherFromPhotonMap(ray, hitInfo, causticMap);
        if (caustic == caustic) { color += caustic; }

        // calculate the number of monte carlo rays to send out for each type
        float diffuseProb, reflectProb, refractProb;
        GetRussianRouletteProbabilities(ray, hitInfo, diffuseProb, reflectProb, refractProb);
        diffuseRaysCount = static_cast<unsigned int>((diffuseProb)* numMonteCarloSamples + 0.49); // round up, but don't want 0 to become 1
        reflectRaysCount = static_cast<unsigned int>((reflectProb)* numMonteCarloSamples + 0.49);
        refractRaysCount = static_cast<unsigned int>((refractProb)* numMonteCarloSamples + 0.49);
    }
    else // specular surfaces
    {
        // do russian roulette to decide between reflection and refraction
        // NOTE: for a tyical surface, only one of the two will be > 0
        if (reflection.Sample(hitInfo.uvw).Gray() + refraction.Sample(hitInfo.uvw).Gray() > 0)
        {
            float diffuseProb, reflectProb, refractProb;
            GetRussianRouletteProbabilities(ray, hitInfo, diffuseProb, reflectProb, refractProb);
            float randVal = RandFloat() * (reflectProb + refractProb);
            if (randVal < reflectProb)
            {
                reflectRaysCount = 1;
            }
            else if (randVal < reflectProb + refractProb)
            {
                refractRaysCount = 1;
            }
        }
    }

    // diffuse
    Color totalIndirect = Color().Black();
    int base1 = 2;
    int base2 = 3;
    int hi = 0;
    for (unsigned int i = 0; i < diffuseRaysCount; ++i)
    {
        Color diffuseColor = Color().Black();
        DiffuseRay(ray, hitInfo, lights, globalMap, causticMap, diffuseColor, base1, base2, hi, bounceCount);
        totalIndirect += diffuseColor * diffuse.Sample(hitInfo.uvw) * diffuseNorm; // explicit no cosine-term since we do importance sampling
        ++hi;
    }
    // reflection
    for (unsigned int i = 0; i < reflectRaysCount; ++i)
    {
        Color reflectColor = Color().Black();
        ReflectRay(ray, hitInfo, lights, globalMap, causticMap, reflectColor, reflectionGlossiness, bounceCount);
        if (reflectionGlossiness > 0) // glossy
        {
            totalIndirect += reflectColor * reflection.Sample(hitInfo.uvw) * ray.dir.Dot(hitInfo.N) * reflectionNorm;
        }
        else // specular
        {
            totalIndirect += reflectColor * reflection.Sample(hitInfo.uvw) * reflectionNorm;
        }
    }
    // refraction
    float reflectRatio = 0;
    float n = ior;
    bool haveRefraction = ComputeRefractParams(ray.dir, hitInfo, n, reflectRatio);
    if (haveRefraction)
    { // trace both rays
        for (unsigned int i = 0; i < refractRaysCount; ++i)
        {
            Color refractColor = Color().Black();
            RefractRay(ray, hitInfo, lights, globalMap, causticMap, refractColor, n, bounceCount);
            Color reflectColor = Color().Black();
            ReflectRay(ray, hitInfo, lights, globalMap, causticMap, reflectColor, refractionGlossiness, bounceCount);
            if (refractionGlossiness > 0) // glossy
            {
                totalIndirect += (refractColor * (1 - reflectRatio) + reflectColor * reflectRatio) * refraction.Sample(hitInfo.uvw) * ray.dir.Dot(hitInfo.N) * refractionNorm;
            }
            else // specular
            {
                totalIndirect += (refractColor * (1 - reflectRatio) + reflectColor * reflectRatio) * refraction.Sample(hitInfo.uvw) * refractionNorm;
            }
        }
    }
    else // total internal reflection
    {
        for (unsigned int i = 0; i < refractRaysCount; ++i)
        {
            Color reflectColor = Color().Black();
            ReflectRay(ray, hitInfo, lights, globalMap, causticMap, reflectColor, refractionGlossiness, bounceCount);
            if (refractionGlossiness > 0) // glossy
            {
                totalIndirect += reflectColor * refraction.Sample(hitInfo.uvw) * ray.dir.Dot(hitInfo.N) * refractionNorm;
            }
            else // specular
            {
                totalIndirect += reflectColor * refraction.Sample(hitInfo.uvw) * refractionNorm;
            }
        }
    }

    auto totalRaysCount = diffuseRaysCount + reflectRaysCount + refractRaysCount;
    if (totalRaysCount > 0)
    {
        totalIndirect /= static_cast<float>(totalRaysCount);
    }
    if (totalIndirect == totalIndirect) { color += totalIndirect; }

    return color;
}

MtlBlinn::TransportType MtlBlinn::DiffuseRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, int base1, int base2, int hi, int bounceCount) const
{
    Ray diffuseRay = RandomRayInHemisphereCosine(hitInfo.p, hitInfo.N);
    //Ray diffuseRay = HaltonRayInHemisphereCosine(hitInfo.p, hitInfo.N, base1, base2, hi);
    HitInfo diffuseHitInfo;
    if (TraceRay(diffuseRay, &rootNode, diffuseHitInfo, HIT_FRONT))
    {
        auto diffuseMat = diffuseHitInfo.node->GetMaterial();
        /* color = diffuseMat->DirectLighting(diffuseRay, diffuseHitInfo, lights);
         if (color.Gray() > 1)
         {
         color /= color.Gray();
         }*/
        if (diffuseMat->IsPhotonSurface())
        {
            color += diffuseMat->GatherFromPhotonMap(diffuseRay, diffuseHitInfo, globalMap);
            color += diffuseMat->GatherFromPhotonMap(diffuseRay, diffuseHitInfo, causticMap);
        }
    }
    else // hit the environment
    {
        color = environment.SampleEnvironment(diffuseRay.dir);
    }
    AbsorbIfNecessary(diffuseRay, diffuseHitInfo, color);
    return DIFFUSE;
}

MtlBlinn::TransportType MtlBlinn::ReflectRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, float glossy, int bounceCount) const
{
    // compute the reflected ray direction
    Ray centerReflectRay;
    centerReflectRay.p = hitInfo.p;
    centerReflectRay.dir = (ray.dir - 2 * (hitInfo.N.Dot(ray.dir) * hitInfo.N)).GetNormalized();
    Ray reflectRay = centerReflectRay;
    TransportType transportType = REFLECT_SPECULAR;

    // decide whether to do mirror or glossy reflection
    if (glossy > 0)
    {
        // glossy reflection
        Point3 c = centerReflectRay.p + centerReflectRay.dir; // center of the glossy sphere
        Point3 q = RandomPointInSphere(c, glossy);
        reflectRay = Ray(centerReflectRay.p, (q - centerReflectRay.p).GetNormalized());
        if (reflectRay.dir.Dot(hitInfo.N) < 0) // ray underneath the plane
        {
            ReflectRayAboutPlane(reflectRay, hitInfo.N);
        }
        transportType = REFLECT_GLOSSY;
        assert(reflectRay.dir.Dot(hitInfo.N) >= 0);
    }

    // shoot the ray
    HitInfo reflectHitInfo;
    if (TraceRay(reflectRay, &rootNode, reflectHitInfo, HIT_FRONT_AND_BACK))
    {
        const Material * mat = reflectHitInfo.node->GetMaterial();
        color = mat->ShadePhoton(reflectRay, reflectHitInfo, lights, globalMap, causticMap, bounceCount + 1);
    }
    else // hit the environment
    {
        color = environment.SampleEnvironment(reflectRay.dir);
    }
    AbsorbIfNecessary(reflectRay, reflectHitInfo, color);
    return transportType;
}

MtlBlinn::TransportType MtlBlinn::RefractRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, float n, int bounceCount) const
{
    TransportType transportType = REFRACT_SPECULAR;

    // compute the refracted ray direction
    Ray centerRefractRay = ComputeRefractionRay(ray.dir.GetNormalized(), hitInfo.p, hitInfo.N, n);
    Ray refractRay = centerRefractRay;
    if (refractionGlossiness > 0) // glossy refraction
    {
        Point3 c = centerRefractRay.p + centerRefractRay.dir; // center of the glossy sphere
        Point3 q = RandomPointInSphere(c, refractionGlossiness);
        refractRay = Ray(centerRefractRay.p, (q - centerRefractRay.p).GetNormalized());
        if (refractRay.dir.Dot(hitInfo.N) > 0) // ray above the surface
        {
            ReflectRayAboutPlane(refractRay, hitInfo.N);
        }
        transportType = REFRACT_GLOSSY;
        assert(refractRay.dir.Dot(hitInfo.N) <= 0);
    }

    // shoot the refraction ray
    HitInfo refractHitInfo;
    if (TraceRay(refractRay, &rootNode, refractHitInfo, HIT_FRONT_AND_BACK))
    {
        const Material * mat = refractHitInfo.node->GetMaterial();
        color = mat->ShadePhoton(refractRay, refractHitInfo, lights, globalMap, causticMap, bounceCount + 1);
    }
    else // hit the environment
    {
        color = environment.SampleEnvironment(refractRay.dir);
    }
    AbsorbIfNecessary(refractRay, refractHitInfo, color);
    return transportType;
}

// ray is incoming ray
Color MtlBlinn::GatherFromPhotonMap(const Ray & ray, const HitInfo & hitInfo, const PhotonMap * map) const
{
    Color color(0, 0, 0);
    if (!hitInfo.front) { return color; }

    static thread_local G3D::Array<Photon> * neighbors = nullptr;
    if (!neighbors) neighbors = new G3D::Array<Photon>();
    float r = radianceEstimationRadius;

    neighbors->resize(0, false);
    G3D::Sphere querySphere(cy2G3D(hitInfo.p), r);
    map->getIntersectingMembers(querySphere, *neighbors);

    float denom = 1 / (PI * Square(r));
    Point3 dirNorm = ray.dir.GetNormalized();
    float f = (glossiness + 2) / 8;
    if (neighbors->size() == 0) return color;
    for (auto & photon : *neighbors)
    {
        if (photon.dir.Dot(dirNorm) > 0) { continue; }
        Color diff = diffuse.Sample(hitInfo.uvw) * diffuseNorm;
        Point3 H = (-dirNorm + photon.dir).GetNormalized();
        float hdotN = H.Dot(hitInfo.N);
        float hdotL = H.Dot(photon.dir);

        if (hdotN > 0 && hdotL > 0)
        {
            Color diff = diffuse.Sample(hitInfo.uvw) * diffuseNorm;
            Color fres = reflection.Sample(hitInfo.uvw) * reflectionNorm * f;
            fres = fres + (1 - fres) * pow(1 - hdotL, 5);
            Color refl = fres * pow(hdotN, glossiness);
            auto c = (diff + refl) * photon.col;
            if (c == c) { color += c; }
            // NOTE: no illumination from the back-face
        }
    }
    color *= denom;
    return color;
}

Color MtlBlinn::DirectLighting(const Ray & ray, const HitInfo & hitInfo, const LightList & lights) const
{
    assert(fabs(hitInfo.N.LengthSquared() - 1) < 1e-4f);
    Color color = Color().Black();

    if (!hitInfo.front) { return color; }

    for (auto & light : lights)
    {
        if (light->IsAmbient())
        {
            color += light->Illuminate(hitInfo.p, hitInfo.N) * diffuse.Sample(hitInfo.uvw);
        }
        else // point or directional
        {
            Point3 L = -light->Direction(hitInfo.p);
            float cosTheta = L.Dot(hitInfo.N);
            if (cosTheta < 0) { continue; }
            // diffuse
            Color illum = light->Illuminate(hitInfo.p, hitInfo.N);
            color += diffuse.Sample(hitInfo.uvw) * cosTheta * illum * diffuseNorm;
            // specular
            Point3 H = (L - ray.dir.GetNormalized()).GetNormalized();
            float hDotN = H.Dot(hitInfo.N);
            float hDotL = H.Dot(L);
            float f = (glossiness + 2) / 8;
            auto fres = specular.Sample(hitInfo.uvw) * reflectionNorm *f;
            fres = fres + (Color(1, 1, 1) - fres) * pow(1 - hDotL, 5);
            color += fres * pow(hDotN, glossiness) * cosTheta * illum;
        }
    }
    return color;
}

Color MtlBlinn::Shade(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, int bounceCount) const
{
    assert(fabs(hitInfo.N.LengthSquared() - 1) < 1e-4f);

    Color color(0, 0, 0);
    Point3 V = (ray.p - hitInfo.p).GetNormalized();
    Point3 N = hitInfo.N;

    // handle lighting
    if (hitInfo.front) // the back-face doesn't receive lighting
    {
        for (auto & light : lights)
        {
            if (light->IsAmbient())
            {
                color += light->Illuminate(hitInfo.p, hitInfo.N) * diffuse.Sample(hitInfo.uvw, hitInfo.duvw);
            }
            else // point or directional
            {
                Point3 L = light->Direction(hitInfo.p) * -1;
                float cosTheta = L.Dot(N);
                if (cosTheta < 0)
                {
                    continue;
                }
                // diffuse
                Color illum = light->Illuminate(hitInfo.p, hitInfo.N);
                color += diffuse.Sample(hitInfo.uvw, hitInfo.duvw) * cosTheta * illum;
                // specular
                Point3 H = (L + V).GetNormalized();
                color += specular.Sample(hitInfo.uvw, hitInfo.duvw) * pow(H.Dot(N), glossiness) * cosTheta * illum;
            }
        }
    }

    if (bounceCount <= 0)
    {
        return color;
    }

    float reflectedRatio = 0; // percentage of light that get reflected
    float cosTheta1 = N.Dot(V);
    float n1 = 1;
    float n2 = 1;

    // handle refraction
    if (refraction.Sample(hitInfo.uvw) != Color().Black())
    {
        n2 = ior;
        if (!hitInfo.front)
        {
            std::swap(n1, n2);
        }

        float sinTheta1 = sqrt(1 - Square(cosTheta1));
        float sinTheta2 = n1 * sinTheta1 / n2;
        if (sinTheta2 > 1) // total internal reflection
        {
            reflectedRatio = 1;
        }

        // calculate the direction of the refracted ray
        if (sinTheta2 < 1) // only in this case we have refracted ray
        {
            float cosTheta2 = sqrt(1 - Square(sinTheta2));
            // compute the percentage of light that gets reflected, using Schlick's approximation to
            // Fresnel equation
            float r0 = Square(n1 - n2) / Square(n1 + n2);
            if (n1 <= n2) // from air to object
            {
                reflectedRatio = r0 + (1 - r0) * pow(1 - cosTheta1, 5);
            }
            else // from object to air
            {
                reflectedRatio = r0 + (1 - r0) * pow(1 - cosTheta2, 5);
            }
            assert(0 <= reflectedRatio && reflectedRatio <= 1);
            if (rand() * 1.0f / RAND_MAX < 1.0f - reflectedRatio) // trace the refracted ray
            {
                // compute the direction of the refracted ray
                Ray centerRefractRay;
                centerRefractRay.p = hitInfo.p;
                Point3 rayDirNormalized = ray.dir.GetNormalized();
                float n = n1 / n2;
                float dDotN = N.Dot(rayDirNormalized);
                float xx = (1 - dDotN * dDotN);
                float yy = n * n * xx;
                float dPrimeDotN = -sqrt(std::max(1 - yy, 0.0f));
                float m = n * dDotN - dPrimeDotN;
                centerRefractRay.dir = n * rayDirNormalized - m * N;
                assert(centerRefractRay.dir.Dot(N) < 0.0001);
                Point3 refractRayNormalized = centerRefractRay.dir.GetNormalized();

                Color totalRefractColor = Color().Black();
                // generate a random point inside the sphere
                float x = 0, y = 0, z = 0;
                do
                {
                    x = (rand() * 1.0f / RAND_MAX * 2 - 1) * refractionGlossiness;
                    y = (rand() * 1.0f / RAND_MAX * 2 - 1) * refractionGlossiness;
                    z = (rand() * 1.0f / RAND_MAX * 2 - 1) * refractionGlossiness;
                } while (x * x + y * y + z * z > refractionGlossiness * refractionGlossiness);

                // compute the ray and reflect it if necessary
                Ray refractRay = centerRefractRay;
                Point3 q = centerRefractRay.p + refractRayNormalized + Point3(x, y, z);
                refractRay.dir = q - centerRefractRay.p;
                if (refractRay.dir.Dot(N) > 0)
                { // reflect
                    q -= N * (2 * refractRay.dir.Dot(N));
                    refractRay.dir = q - refractRay.p;
                }

                // shoot the refracted ray
                HitInfo refractHitInfo;
                bool refractHit = TraceRay(refractRay, &rootNode, refractHitInfo, HIT_FRONT_AND_BACK);
                if (refractHit) // the refraction ray hits something
                {
                    const Material * refractMat = refractHitInfo.node->GetMaterial();
                    Color refractColor = refractMat->Shade(refractRay, refractHitInfo, lights, bounceCount - 1);
                    if (n2 == 1) // refracted ray travels in air, no absorption
                    {
                        totalRefractColor += refractColor * refraction.Sample(hitInfo.uvw) * (1 - reflectedRatio);
                    }
                    else // refracted ray travels inside the object, gets absorbed
                    {
                        Point3 travelVec = refractHitInfo.p - refractRay.p;
                        float d = travelVec.Length();
                        totalRefractColor.r += refractColor.r * refraction.Sample(hitInfo.uvw).r * exp(-d * absorption.r) * (1 - reflectedRatio);
                        totalRefractColor.g += refractColor.g * refraction.Sample(hitInfo.uvw).g * exp(-d * absorption.g) * (1 - reflectedRatio);
                        totalRefractColor.b += refractColor.b * refraction.Sample(hitInfo.uvw).b * exp(-d * absorption.b) * (1 - reflectedRatio);
                    }
                }
                else // the refracted ray hits nothing
                {
                    // sample the environment
                    totalRefractColor += refraction.Sample(hitInfo.uvw) * environment.SampleEnvironment(refractRay.dir.GetNormalized());
                }
                color += totalRefractColor;
            }
        }
    } // end refraction

    // handle reflection
    if (reflection.Sample(hitInfo.uvw) != Color().Black() || (reflectedRatio > 0 && rand() * 1.0f / RAND_MAX < reflectedRatio))
    {
        // calculate the direction of the reflected ray
        Ray centerReflectRay;
        centerReflectRay.p = hitInfo.p;
        centerReflectRay.dir = ray.dir - 2 * (N.Dot(ray.dir) * N);

        // shoot multiple rays for glossy reflection
        Point3 reflectRayNormalized = centerReflectRay.dir.GetNormalized();
        Color totalReflectColor = Color().Black();

        // generate a random point inside the sphere
        float x = 0, y = 0, z = 0;
        do
        {
            x = (rand() * 1.0f / RAND_MAX * 2 - 1) * reflectionGlossiness;
            y = (rand() * 1.0f / RAND_MAX * 2 - 1) * reflectionGlossiness;
            z = (rand() * 1.0f / RAND_MAX * 2 - 1) * reflectionGlossiness;
        } while (x * x + y * y + z * z > reflectionGlossiness * reflectionGlossiness);

        // compute the ray and reflect it if necessary
        Ray reflectRay = centerReflectRay;
        Point3 q = centerReflectRay.p + reflectRayNormalized + Point3(x, y, z);
        reflectRay.dir = q - centerReflectRay.p;
        if (reflectRay.dir.Dot(N) < 0)
        { // reflect
            q += N * (2 * -reflectRay.dir.Dot(N));
            reflectRay.dir = q - reflectRay.p;
        }

        // shoot the reflected ray
        HitInfo reflectHitInfo;
        bool reflectHit = TraceRay(reflectRay, &rootNode, reflectHitInfo, HIT_FRONT_AND_BACK);
        if (reflectHit) // the reflected ray hits some object
        {
            const Material * reflectMat = reflectHitInfo.node->GetMaterial();
            Color reflectColor = reflectMat->Shade(reflectRay, reflectHitInfo, lights, bounceCount - 1);
            Point3 travelVec = reflectHitInfo.p - reflectRay.p;
            if (reflection.Sample(hitInfo.uvw) != Color().Black()) // pure reflection
            {
                totalReflectColor += reflectColor * reflection.Sample(hitInfo.uvw);
            }
            if (reflectedRatio > 0) // reflection due to refraction
            {
                if (n1 == 1) // reflected ray travels in air, no absorption
                {
                    totalReflectColor += reflectColor * refraction.Sample(hitInfo.uvw) * reflectedRatio;
                }
                else // reflected ray travels inside the object, gets absorbed
                {
                    Point3 travelVec = reflectHitInfo.p - reflectRay.p;
                    float d = travelVec.Length();
                    totalReflectColor.r += reflectColor.r * refraction.Sample(hitInfo.uvw).r * exp(-d * absorption.r) * reflectedRatio;
                    totalReflectColor.g += reflectColor.g * refraction.Sample(hitInfo.uvw).g * exp(-d * absorption.g) * reflectedRatio;
                    totalReflectColor.b += reflectColor.b * refraction.Sample(hitInfo.uvw).b * exp(-d * absorption.b) * reflectedRatio;
                }
            }
        }
        else // the reflected ray hit nothing
        {
            // sample the environment
            if (reflection.Sample(hitInfo.uvw) != Color().Black()) // pure reflection
            {
                totalReflectColor += reflection.Sample(hitInfo.uvw) * environment.SampleEnvironment(reflectRay.dir.GetNormalized());
            }
            if (reflectedRatio > 0) // reflection due to refraction
            {
                totalReflectColor += refraction.Sample(hitInfo.uvw) * environment.SampleEnvironment(reflectRay.dir.GetNormalized()) * reflectedRatio;
            }
        }
        color += totalReflectColor;
    } // end reflection

    // handle diffuse surfaces
    if (diffuse.Sample(hitInfo.uvw) != Color().Black() && bounceCount > 0)
    {
        // find two vectors perpendicular to the normal
        // cross product the normal with the x axis
        Point3 axis1 = N.Cross(Point3(1, 0, 0)).GetNormalized();
        if (axis1.LengthSquared() < 1e-3)
        {
            axis1 = N.Cross(Point3(0, 1, 0)).GetNormalized();
        }
        // cross product the previous two vectors to get the other axis
        Point3 axis2 = N.Cross(axis1).GetNormalized();

        Color totalDiffColor = Color().Black();

        // generate random points on the disk
        float r = sqrt(rand() * 1.0f / RAND_MAX);
        float phi = TWO_PI * rand() * 1.0f / RAND_MAX;
        // project these points to the hemisphere
        Point3 x = axis1 * r * cos(phi);
        Point3 y = axis2 * r * sin(phi);
        Point3 z = N * sqrt(1 - r * r);
        Point3 q = (x + y + z).GetNormalized();
        // send a ray from p to q
        HitInfo diffHitInfo;
        Ray diffRay(hitInfo.p, q);
        bool diffHit = TraceRay(diffRay, &rootNode, diffHitInfo, HIT_FRONT_AND_BACK);
        if (diffHit)
        {
            const Material * diffMat = diffHitInfo.node->GetMaterial();
            Color diffColor = diffMat->Shade(diffRay, diffHitInfo, lights, bounceCount - 1);
            totalDiffColor += diffColor * diffuse.Sample(hitInfo.uvw);
        }
        else
        {
            totalDiffColor += diffuse.Sample(hitInfo.uvw) * environment.SampleEnvironment(diffRay.dir.GetNormalized());
        }
        color += totalDiffColor;
    }

    return color;
}

Material::TransportType MtlBlinn::RandomPhotonBounce(Ray & photonRay, Color & photonColor, const HitInfo & hitInfo) const
{
    Color factor = Color().White();
    // check absorption
    AbsorbIfNecessary(photonRay, hitInfo, factor);

    float diffuseProb, reflectProb, refractProb;

    GetRussianRouletteProbabilities(photonRay, hitInfo, diffuseProb, reflectProb, refractProb);
    float randVal = RandFloat();

    // compute the transport type and ray using russian roulette
    TransportType transportType = ABSORB;
    if (randVal < diffuseProb)
    {
        factor *= diffuse.Sample(hitInfo.uvw) * diffuseNorm / diffuseProb;
        DiffusePhoton(photonRay, hitInfo);
        transportType = DIFFUSE;
    }
    else if (randVal < diffuseProb + reflectProb)
    {
        factor *= (1 / reflectProb) * reflection.Sample(hitInfo.uvw) * reflectionNorm;
        ReflectPhoton(photonRay, hitInfo, reflectionGlossiness);
        if (reflectionGlossiness > 0)
        {
            transportType = REFLECT_GLOSSY;
            factor *= photonRay.dir.Dot(hitInfo.N);
        }
        else
        {
            transportType = REFLECT_SPECULAR;
        }
    }
    else if (randVal < diffuseProb + reflectProb + refractProb)
    {
        factor *= (1 / refractProb) * refraction.Sample(hitInfo.uvw) * refractionNorm;
        float reflectRatio = 0;
        float n = ior;
        bool haveRefraction = ComputeRefractParams(photonRay.dir.GetNormalized(), hitInfo, n, reflectRatio);
        if (haveRefraction || RandFloat() < (1 - reflectRatio)) // refraction
        {
            RefractPhoton(photonRay, hitInfo, n);
            if (refractionGlossiness > 0)
            {
                transportType = REFRACT_GLOSSY;
                factor *= -photonRay.dir.Dot(hitInfo.N);
            }
            else
            {
                transportType = REFRACT_SPECULAR;
            }
        }
        else // fresnel reflection or total internal reflection
        {
            ReflectPhoton(photonRay, hitInfo, refractionGlossiness);
            if (refractionGlossiness > 0)
            {
                transportType = REFLECT_GLOSSY_FRESNEL;
                factor *= photonRay.dir.Dot(hitInfo.N);
            }
            else
            {
                transportType = REFLECT_SPECULAR_FRESNEL;
            }
        }
    }

    assert(factor.Gray() > 0);
    photonColor *= factor;
    return transportType;
}

void MtlBlinn::DiffusePhoton(Ray & photonRay, const HitInfo & hitInfo) const
{
    // select a random cosine-weighted direction above the hemisphere
    photonRay = RandomRayInHemisphereCosine(hitInfo.p, hitInfo.N);
}

// Returns true if the reflection is glossy, and false if the reflection is specular
void MtlBlinn::ReflectPhoton(Ray & photonRay, const HitInfo & hitInfo, float glossy) const
{
    // compute the reflected ray direction
    Ray centerReflectRay;
    centerReflectRay.p = hitInfo.p;
    centerReflectRay.dir = photonRay.dir - 2 * (hitInfo.N.Dot(photonRay.dir) * hitInfo.N);
    centerReflectRay.dir.Normalize();

    // decide whether to do mirror or glossy reflection
    if (glossy > 0)
    {
        // glossy reflection
        Point3 c = centerReflectRay.p + centerReflectRay.dir; // center of the glossy sphere
        Point3 q = RandomPointInSphere(c, glossy);
        photonRay = Ray(centerReflectRay.p, (q - centerReflectRay.p).GetNormalized());
        if (photonRay.dir.Dot(hitInfo.N) < 0) // ray underneath the plane
        {
            ReflectRayAboutPlane(photonRay, hitInfo.N);
        }
        assert(photonRay.dir.Dot(hitInfo.N) >= 0);
    }
    // specular reflection
    photonRay = centerReflectRay;
}

// incoming ray should be normalized
// this returns false if the surface has no refraction or total internal reflection happens
bool MtlBlinn::ComputeRefractParams(const Point3 & incomingRay, const HitInfo & hitInfo, float & n, float & reflectRatio) const
{
    float cosTheta1 = hitInfo.N.Dot(-incomingRay); assert(cosTheta1 >= 0 && cosTheta1 <= 1);
    float n1 = hitInfo.front ? 1 : ior;
    float n2 = hitInfo.front ? ior : 1;
    n = n1 / n2;
    float sinTheta1 = sqrt(std::max(1 - Square(cosTheta1), 0.0f));
    float sinTheta2 = n1 * sinTheta1 / n2;
    bool haveRefraction = refraction.Sample(hitInfo.uvw).Gray() > 0;
    if (sinTheta2 >= 1) // total internal reflection
    {
        reflectRatio = 1;
        haveRefraction = false;
    }
    else // we have refraction
    {
        // compute the ratio of light that get reflected
        float r0 = Square(n1 - n2) / Square(n1 + n2);
        reflectRatio = (n1 <= n2) ? (r0 + (1 - r0) * pow(1 - cosTheta1, 5)) : 0; // from air to object
        // NOTE: we disallow Fresnel reflection for back face
        assert(reflectRatio <= 1);
        haveRefraction = true;
    }
    return haveRefraction;
}

void MtlBlinn::RefractPhoton(Ray & photonRay, const HitInfo & hitInfo, float n) const
{
    // compute the refracted ray direction
    photonRay.dir.Normalize();
    Ray centerRefractRay = ComputeRefractionRay(photonRay.dir, hitInfo.p, hitInfo.N, n);
    if (refractionGlossiness > 0) // glossy refraction
    {
        Point3 c = centerRefractRay.p + centerRefractRay.dir; // center of the glossy sphere
        Point3 q = RandomPointInSphere(c, refractionGlossiness);
        photonRay = Ray(centerRefractRay.p, (q - centerRefractRay.p).GetNormalized());
        if (photonRay.dir.Dot(hitInfo.N) > 0) // ray above the surface
        {
            ReflectRayAboutPlane(photonRay, hitInfo.N);
        }
        assert(photonRay.dir.Dot(hitInfo.N) <= 0);
    }
    // specular refraction
    photonRay = centerRefractRay;
}

void MtlBlinn::SetViewportMaterial(int subMtlID) const
{
    (void*)subMtlID;
    // empty
}