//-------------------------------------------------------------------------------
///
/// \file       materials.h
/// \author     Cem Yuksel (www.cemyuksel.com)
/// \version    12.3
/// \date       December 4, 2014
///
/// \brief Example source for CS 6620 - University of Utah.
///
//-------------------------------------------------------------------------------

#ifndef _MATERIALS_H_INCLUDED_
#define _MATERIALS_H_INCLUDED_

#include "scene.h"

//-------------------------------------------------------------------------------

class MtlBlinn : public Material
{
public:
    MtlBlinn() : diffuse(0.5f, 0.5f, 0.5f), specular(0.7f, 0.7f, 0.7f), glossiness(20.0f), emission(0, 0, 0),
        reflection(0, 0, 0), refraction(0, 0, 0), absorption(0, 0, 0), ior(1), viewportTextureID(0),
        reflectionGlossiness(0), refractionGlossiness(0), diffuseNorm(1), reflectionNorm(1), refractionNorm(1) {}
    virtual Color Shade(const Ray &ray, const HitInfo &hInfo, const LightList &lights, int bounceCount) const;
    virtual Color ShadePhoton(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, int bounceCount) const;
    virtual Color ShadePhotonDirect(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap) const;

    virtual void Normalize();

    void SetDiffuse(Color dif) { diffuse.SetColor(dif); }
    void SetSpecular(Color spec) { specular.SetColor(spec); }
    void SetGlossiness(float gloss) { glossiness = gloss; }
    void SetEmission(Color e)       { emission.SetColor(e); }
    void SetReflection(Color reflect) { reflection.SetColor(reflect); }
    void SetRefraction(Color refract) { refraction.SetColor(refract); }
    void SetAbsorption(Color absorp) { absorption = absorp; }
    void SetRefractionIndex(float _ior) { ior = _ior; }

    void SetDiffuseTexture(TextureMap *map) { diffuse.SetTexture(map); }
    void SetSpecularTexture(TextureMap *map) { specular.SetTexture(map); }
    void SetEmissionTexture(TextureMap *map)  { emission.SetTexture(map); }
    void SetReflectionTexture(TextureMap *map) { reflection.SetTexture(map); }
    void SetRefractionTexture(TextureMap *map) { refraction.SetTexture(map); }
    void SetReflectionGlossiness(float gloss) { reflectionGlossiness = gloss; }
    void SetRefractionGlossiness(float gloss) { refractionGlossiness = gloss; }

    virtual void SetViewportMaterial(int subMtlID = 0) const;   // used for OpenGL display

    // Photon Extensions
    virtual bool IsPhotonSurface() const { return diffuse.GetColor().Gray() > 0; }   // if this method returns true, the photon will be stored
    virtual TransportType RandomPhotonBounce(Ray &r, Color &c, const HitInfo &hInfo) const;  // if this method returns true, a new photon with the given direction and color will be traced
    virtual Color DirectLighting(const Ray & ray, const HitInfo & hitInfo, const LightList & lights) const;
    virtual Color GatherFromPhotonMap(const Ray & ray, const HitInfo & hitInfo, const PhotonMap * map) const;

private:
    TexturedColor diffuse, specular, reflection, refraction, emission;
    float glossiness;
    Color absorption;
    float ior;  // index of refraction
    float reflectionGlossiness, refractionGlossiness;
    unsigned int viewportTextureID;
    float diffuseNorm;
    float reflectionNorm;
    float refractionNorm;
    void DiffusePhoton(Ray & photonRay, const HitInfo & hitInfo) const;
    void ReflectPhoton(Ray & photonRay, const HitInfo & hitInfo, float glossy) const;
    void RefractPhoton(Ray & photonRay, const HitInfo & hitInfo, float n) const;
    bool ComputeRefractParams(const Point3 & incomingRay, const HitInfo & hitInfo, float & n, float & reflectRatio) const;
    TransportType DiffuseRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, int base1, int base2, int hi, int bounceCount) const;
    TransportType ReflectRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, float glossy, int bounceCount) const;
    TransportType RefractRay(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, Color & color, float n, int bounceCount) const;
    void AbsorbIfNecessary(const Ray & ray, const HitInfo & hitInfo, Color & color) const
    {
        if (!hitInfo.front && ior > 1)
        {
            float d = (ray.p - hitInfo.p).Length();
            color.r *= exp(-d * absorption.r);
            color.g *= exp(-d * absorption.g);
            color.b *= exp(-d * absorption.b);
        }
    }

    void GetRussianRouletteProbabilities(const Ray & ray, const HitInfo & hitInfo, float & diffuseProb, float & reflectProb, float & refractProb) const
    {
        float diffuseGray = diffuse.Sample(hitInfo.uvw).Gray() * diffuseNorm;
        float reflectGray = reflection.Sample(hitInfo.uvw).Gray() * reflectionNorm;
        float refractGray = refraction.Sample(hitInfo.uvw).Gray() * refractionNorm;
        float totalGray = diffuseGray + reflectGray + refractGray;
        diffuseProb = diffuseGray / totalGray;
        reflectProb = reflectGray / totalGray;
        refractProb = refractGray / totalGray;
    }
};

//-------------------------------------------------------------------------------

class MultiMtl : public Material
{
public:
    virtual ~MultiMtl() { for (unsigned int i = 0; i < mtls.size(); i++) delete mtls[i]; }

    virtual Color Shade(const Ray &ray, const HitInfo &hInfo, const LightList &lights, int bounceCount) const { return hInfo.mtlID < (int)mtls.size() ? mtls[hInfo.mtlID]->Shade(ray, hInfo, lights, bounceCount) : Color(1, 1, 1); }
    virtual Color ShadePhoton(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap, int bounceCount) const
    {
        if (hitInfo.mtlID < (int)mtls.size())
            return mtls[hitInfo.mtlID]->ShadePhoton(ray, hitInfo, lights, globalMap, causticMap, bounceCount);
        return Color(1, 1, 1);
    }
    virtual Color ShadePhotonDirect(const Ray & ray, const HitInfo & hitInfo, const LightList & lights, const PhotonMap * globalMap, const PhotonMap * causticMap) const
    {
        if (hitInfo.mtlID < (int)mtls.size())
            return mtls[hitInfo.mtlID]->ShadePhotonDirect(ray, hitInfo, lights, globalMap, causticMap);
        return Color(1, 1, 1);
    }
    virtual void Normalize() { for (auto & mtl : mtls) { mtl->Normalize(); } };

    virtual void SetViewportMaterial(int subMtlID = 0) const { if (subMtlID < (int)mtls.size()) mtls[subMtlID]->SetViewportMaterial(); }

    // Photon Extensions
    virtual bool IsPhotonSurface() const { for (unsigned int i = 0; i < mtls.size(); i++) if (mtls[i]->IsPhotonSurface()) return true; return false; }
    virtual TransportType RandomPhotonBounce(Ray &r, Color &c, const HitInfo &hInfo) const { return hInfo.mtlID < (int)mtls.size() ? mtls[hInfo.mtlID]->RandomPhotonBounce(r, c, hInfo) : ABSORB; }
    virtual Color GatherFromPhotonMap(const Ray & ray, const HitInfo & hitInfo, const PhotonMap * map) const
    {
        if (hitInfo.mtlID < (int)mtls.size())
        {
            return mtls[hitInfo.mtlID]->GatherFromPhotonMap(ray, hitInfo, map);
        }
        return Color(0, 0, 0);
    }

    virtual Color DirectLighting(const Ray & ray, const HitInfo & hitInfo, const LightList & lights) const
    {
        if (hitInfo.mtlID < (int)mtls.size())
        {
            return mtls[hitInfo.mtlID]->DirectLighting(ray, hitInfo, lights);
        }
        return Color(0, 0, 0);
    }
    void AppendMaterial(Material *m) { mtls.push_back(m); }

private:
    std::vector<Material*> mtls;
};

//-------------------------------------------------------------------------------

#endif
