#include <objects.h>
#include <common.h>

void Sphere::ViewportDisplay(const Material *mtl) const
{
    (void*)mtl;
    // empty
}

void TriObj::ViewportDisplay(const Material *mtl) const
{
    (void*)mtl;
    // empty
}

void Plane::ViewportDisplay(const Material *mtl) const
{
    (void*)mtl;
    // empty
}

static const float parallelEps = 1e-3f;
static const float hitBias = 5e-3f;

// assuming this ray has been transformed to the coordinate frame of the sphere
bool Sphere::IntersectRay(const Ray & ray, HitInfo & hitInfo, int hitSide) const
{
    float a = ray.dir.Dot(ray.dir);
    float b = 2 * ray.p.Dot(ray.dir);
    float c = ray.p.Dot(ray.p) - 1;
    float disc = b * b - 4 * a * c;

    if (disc < 0 || a == 0)
    {
        return false;
    }

    float discSqrt = sqrtf(disc);

    float t0 = (-b - discSqrt) / (2 * a);
    float t1 = (-b + discSqrt) / (2 * a);

    if (t0 > t1)
    {
        std::swap(t0, t1);
    }

    // avoid self-hit
    if (t1 <= hitBias)
    {
        return false;
    }

    // check if there is already a nearer hit
    float z = t0 <= hitBias ? t1 : t0;
    if (z >= hitInfo.z)
    {
        return false;
    }

    // check if we hit the correct face
    cyPoint3f p = ray.p + z * ray.dir; // hit point
    cyPoint3f N = p; // normal
    bool front = N.Dot(-ray.dir) >= 0;
    if ((front && hitSide == HIT_BACK) || (!front && hitSide == HIT_FRONT))
    {
        return false;
    }

    // we have found a hit
    hitInfo.z = z;
    hitInfo.p = p;
    hitInfo.front = front;
    hitInfo.N = hitInfo.front ? N : -N;

    // do a spherical mapping
    hitInfo.uvw = cyPoint3f((atan2(p.y, p.x) + PI) / TWO_PI, (asin(-p.z) + HALF_PI) / PI, 0.0f);
    return true;
}

// N is one of the three unit vectors along x, y and z
inline float IntersectRayPlane(const Ray & ray, float x, const cyPoint3f & N)
{
    // check if the ray is parallel to the plane
    float denom = ray.dir.Dot(N);
    if (fabs(denom) <= parallelEps)
    {
        return FLT_MAX;
    }

    return (x - N.Dot(ray.p)) / denom;
}

bool Box::IntersectRay(const Ray & ray, float t_max) const
{
    // X planes
    float tx1 = IntersectRayPlane(ray, pmin.x, cyPoint3f(1.0f, 0.0f, 0.0f));
    float tx2 = IntersectRayPlane(ray, pmax.x, cyPoint3f(1.0f, 0.0f, 0.0f));
    if (tx1 > tx2)
    {
        std::swap(tx1, tx2);
    }
    if (tx1 == FLT_MAX) // parallel case
    {
        tx1 = -FLT_MAX;
        if (ray.p.x < pmin.x || ray.p.x > pmax.x)
        {
            return false;
        }
    }

    // Y planes
    float ty1 = IntersectRayPlane(ray, pmin.y, cyPoint3f(0.0f, 1.0f, 0.0f));
    float ty2 = IntersectRayPlane(ray, pmax.y, cyPoint3f(0.0f, 1.0f, 0.0f));
    if (ty1 > ty2)
    {
        std::swap(ty1, ty2);
    }
    if (ty1 == FLT_MAX) // parallel case
    {
        ty1 = -FLT_MAX;
        if (ray.p.y < pmin.y || ray.p.y > pmax.y)
        {
            return false;
        }
    }

    // Z planes
    float tz1 = IntersectRayPlane(ray, pmin.z, cyPoint3f(0.0f, 0.0f, 1.0f));
    float tz2 = IntersectRayPlane(ray, pmax.z, cyPoint3f(0.0f, 0.0f, 1.0f));
    if (tz1 > tz2)
    {
        std::swap(tz1, tz2);
    }
    if (tz1 == FLT_MAX) // parallel case
    {
        tz1 = -FLT_MAX;
        if (ray.p.z < pmin.z || ray.p.z > pmax.z)
        {
            return false;
        }
    }

    // find the range
    float t1 = std::max(std::max(tx1, ty1), tz1);
    float t2 = std::min(std::min(tx2, ty2), tz2);

    return t1 <= t2 && t1 <= t_max;
}

// Intersect a ray with the unit plane
bool Plane::IntersectRay(const Ray & ray, HitInfo & hitInfo, int hitSide) const
{
    // check to see if we hit the correct face
    bool front = ray.dir.z <= 0;
    if ((front && hitSide == HIT_BACK) || (!front && hitSide == HIT_FRONT))
    {
        return false;
    }

    // check if the ray is parallel to the plane
    if (fabs(ray.dir.z) <= parallelEps)
    {
        return false;
    }

    // check if there is a closer hit
    float t = -ray.p.z / ray.dir.z;
    if (t >= hitInfo.z || t <= hitBias)
    {
        return false;
    }

    // check if the point is inside the unit plane
    cyPoint3f p = ray.p + ray.dir * t;
    p.z = 0;
    if (fabs(p.x) > 1 || fabs(p.y) > 1)
    {
        return false;
    }

    // we have found a hit
    hitInfo.front = front;
    hitInfo.N = hitInfo.front ? cyPoint3f(0.0f, 0.0f, 1.0f) : cyPoint3f(0.0f, 0.0f, -1.0f);
    hitInfo.p = p;
    hitInfo.z = t;
    hitInfo.uvw = cyPoint3f(0.5f * (p.x + 1.0f), 0.5f * (p.y + 1.0f), 0.0f);

    return true;
}

// Intersect a ray with a triangle mesh
bool TriObj::IntersectRay(const Ray & ray, HitInfo & hitInfo, int hitSide) const
{
    // loop through the triangles and intersect with each of them until we find a hit
    bool hit = false;
    hit = TraceBVHNode(ray, hitInfo, hitSide, bvh.GetRootNodeID());
    return hit;
}

// Intersect a ray with a triangle
bool TriObj::IntersectTriangle(const Ray & ray, HitInfo & hitInfo, int hitSide, unsigned int faceID) const
{
    // get the face
    const cyTriFace & face = F(faceID);

    // get the three vertices
    const cyPoint3f & v0 = V(face.v[0]);
    const cyPoint3f & v1 = V(face.v[1]);
    const cyPoint3f & v2 = V(face.v[2]);

    // get the true normal of the face
    cyPoint3f N = (v1 - v0).Cross(v2 - v0).GetNormalized();

    // check to see if the ray is parallel to the face
    const cyPoint3f & r = ray.p;
    const cyPoint3f & d = ray.dir;
    float denom = N.Dot(d);
    if (fabs(denom) <= parallelEps)
    {
        return false;
    }

    // get the intersection
    float t = N.Dot(v0 - r) / N.Dot(d);
    cyPoint3f p = r + d * t;

    // check if we hit the correct face
    bool front = N.Dot(-d) >= 0;
    if ((front && hitSide == HIT_BACK) || (!front && hitSide == HIT_FRONT))
    {
        return false;
    }

    // check the t value
    if (t <= hitBias || t >= hitInfo.z)
    {
        return false;
    }

    // check to see if this point is inside the triangle
    const float eps = 1e-5f;
    const float minBound = -eps;
    const float maxBound = 1.0f + eps;
    cyPoint3f e1 = v1 - v0;
    cyPoint3f e2 = v2 - v0;
    float A = 0.5f * (e1.Cross(e2)).Dot(N); // total area of the triangle
    float A2 = 0.5f * (e1.Cross(p - v0)).Dot(N);
    float w2 = A2 / A; // barycentric coordinate of p with regard to v2
    if (w2 < minBound || w2 > maxBound)
    {
        return false;
    }
    float A1 = 0.5f * ((p - v0).Cross(e2)).Dot(N);
    float w1 = A1 / A; // barycentric coordinate of p with regard to v1
    if (w1 < minBound || w1 > maxBound)
    {
        return false;
    }
    float w0 = 1.0f - w1 - w2;
    if (w0 < minBound || w0 > maxBound)
    {
        return false;
    }
    cyPoint3f baryCentricCoords(w0, w1, w2);

    // we have found a hit
    hitInfo.z = t;
    hitInfo.p = p;
    hitInfo.front = front;
    cyPoint3f n = hitInfo.front ? GetNormal(faceID, baryCentricCoords) : -GetNormal(faceID, baryCentricCoords);
    hitInfo.N = n.GetNormalized();
    hitInfo.mtlID = GetMaterialIndex(faceID);
    if (HasTextureVertices())
    {
        hitInfo.uvw = GetTexCoord(faceID, baryCentricCoords);
    }
    return true;
}

bool TriObj::TraceBVHNode(const Ray & ray, HitInfo & hitInfo, int hitSide, unsigned int nodeID) const
{
    // check the ray against the bounding box of the node
    // if it does not hit the bounding box, return false
    const float * points = bvh.GetNodeBounds(nodeID);
    Box bbox(points[0], points[1], points[2], points[3], points[4], points[5]);
    if (!bbox.IntersectRay(ray, hitInfo.z))
    {
        return false;
    }

    bool hit = false;

    // if the node is leaf
    if (bvh.IsLeafNode(nodeID))
    {
        // get the list of triangles
        unsigned int numElems = bvh.GetNodeElementCount(nodeID);
        const unsigned int * elems = bvh.GetNodeElements(nodeID);
        // check the ray against each triangle
        for (unsigned int i = 0; i < numElems; ++i)
        {
            if (IntersectTriangle(ray, hitInfo, hitSide, elems[i]))
            {
                hit = true;
                if (hitInfo.shadow) return true;
            }
        }
    }
    else // internal node
    {
        unsigned int firstChild = bvh.GetFirstChildNode(nodeID);
        bool firstHit = TraceBVHNode(ray, hitInfo, hitSide, firstChild);
        if (firstHit && hitInfo.shadow) return true;
        unsigned int secondChild = bvh.GetSecondChildNode(nodeID);
        bool secondHit = TraceBVHNode(ray, hitInfo, hitSide, secondChild);
        hit = firstHit || secondHit;
    }

    return hit;
}